## Helping formating classes for to_c

class   indentable:
    """
    base of all fmt objects
    """
    char_indent = " "
    num_indent = 4

    def __init__(self):
        self._indent = 0
        self._is_indented = False
        self._lsdata = None

    def to_str(self, res: str) -> str:
        pass

    def __str__(self):
        self.set_indent()
        strinit = (self.char_indent * self.num_indent) * (self._indent - 1)
        return self.to_str(strinit, self._indent)

    @property
    def lsdata(self) -> list:
        return self._lsdata

    def set_indent(self, indent: int=1):
        if self._is_indented:
            return
        self._indent = indent
        if isinstance(self._lsdata, indentable):
            self._lsdata.set_indent(self._indent)
        if isinstance(self._lsdata, list):
            list_set_indent(self._lsdata, self._indent)
        self._is_indented = True
    
def catend(dst :str, src :str, indent) -> str:
    """cat two strings but handle \n for tabulation"""
    res = dst
    txtsrc = src
    if not isinstance(src, str):
        txtsrc = str(src)
    for c in list(txtsrc):
        if len(res) > 0 and res[-1] == '\n':
            res += (indentable.char_indent * indentable.num_indent) * (indent - 1) + c
        else:
            res += c
    return res

def list_set_indent(lst: list, indent: int=1):
    """recurs into list for indentation"""
    for i in lst:
        if isinstance(i, indentable):
            i.set_indent(indent)
        if isinstance(i, list):
            list_set_indent(i, indent)

def list_to_str(lst: list, content: str, indent: int=1):
    """recurs into list for string computing """
    for i in lst:
        if isinstance(i, indentable):
            content = i.to_str(content, indent)
        elif isinstance(i, list):
            content = list_to_str(i, content, indent)
        elif isinstance(i, str):
            content = catend(content, i, indent)
    return content

class   block(indentable):
    """
    for {}, (), []
    """
    def __init__(self, beginby :str, endby :str, lsdata :list):
        indentable.__init__(self)
        self._beginby = beginby
        self._endby = endby
        if lsdata == None:
            raise Exception("lsdata can't be None")
        self._lsdata = lsdata

    def to_str(self, res: str, parent_indent) -> str:
        self.set_indent()
        content = catend(res, self._beginby, parent_indent)
        if isinstance(self._lsdata, indentable):
            return catend(self._lsdata.to_str(content, self._indent), self._endby, parent_indent)
        if isinstance(self._lsdata, list):
            content = list_to_str(self._lsdata, content, self._indent)
        return catend(content, self._endby, parent_indent)

class   sep(indentable):
    """
    for all list seperated by a char
    """
    def __init__(self, ch :str, lsdata :list):
        indentable.__init__(self)
        self._ch = ch
        if lsdata == None:
            raise Exception("lsdata can't be None")
        self._lsdata = lsdata

    def to_str(self, res: str, parent_indent) -> str:
        self.set_indent()
        content = res
        if isinstance(self._lsdata, indentable):
            return self._lsdata.to_str(res, self._indent)
        if isinstance(self._lsdata, list):
            sz = len(self._lsdata)
            for i in range(sz):
                if isinstance(self._lsdata[i], indentable):
                    content = self._lsdata[i].to_str(content, self._indent)
                elif isinstance(self._lsdata[i], list):
                    content = list_to_str(self._lsdata[i], content, self._indent)
                else:
                    content = catend(content, self._lsdata[i], self._indent)
                if i < sz - 1:
                    content = catend(content, self._ch, self._indent)
        return content

class   end(indentable):
    """
    for all list that end by a char
    """
    def __init__(self, ch :str, lsdata :list):
        indentable.__init__(self)
        self._ch = ch
        if lsdata == None:
            raise Exception("lsdata can't be None")
        self._lsdata = lsdata

    def to_str(self, res: str, parent_indent) -> str:
        self.set_indent()
        content = res
        if isinstance(self._lsdata, indentable):
            return catend(self._lsdata.to_str(res, self._indent), self._ch, self._indent)
        if isinstance(self._lsdata, list):
            for i in self._lsdata:
                if isinstance(i, indentable):
                    content = i.to_str(content, self._indent)
                elif isinstance(i, list):
                    content = list_to_str(i, content, self._indent)
                else:
                    content = catend(content, i, self._indent)
                content = catend(content, self._ch, self._indent)
        return content

class   tab(indentable):
    """
    to handle indentation level
    """
    def __init__(self, lsdata: indentable):
        indentable.__init__(self)
        if lsdata == None:
            raise Exception("lsdata can't be None")
        self._lsdata = lsdata

    def set_indent(self, indent: int=1):
        if self._is_indented:
            return
        self._indent = indent + 1
        if isinstance(self._lsdata, indentable):
            self._lsdata.set_indent(self._indent)
        if isinstance(self._lsdata, list):
            list_set_indent(self._lsdata, self._indent)
        self._is_indented = True

    def to_str(self, res: str, parent_indent) -> str:
        self.set_indent()
        if isinstance(self._lsdata, indentable):
            return self._lsdata.to_str(res, self._indent)
        if isinstance(self._lsdata, list):
            content = res
            for i in self._lsdata:
                if isinstance(i, indentable):
                    content = i.to_str(content, self._indent)
                elif isinstance(i, list):
                    content = list_to_str(i, content, self._indent)
                else:
                    content = catend(content, i, self._indent)
        return content



from pyrser import meta
from pyrser import error
from pyrser.parsing.parserStream import Stream
from pyrser.parsing.node import Node
import collections

# TODO: ensure unicity of names
#: Module variable to store meta class instance by classname
_MetaBasicParser = {}


class MetaBasicParser(type):
    """Metaclass for all parser."""
    def __new__(metacls, name, bases, namespace):
        global _MetaBasicParser
        # create the metaclass instance
        cls = type.__new__(metacls, name, bases, namespace)
        # search metaclass instance of all base
        if len(bases) > 1:
            raise TypeError("%s must inherit from an unique parent,"
                            " use Grammar for aggregation" % name)
        # Manage inheritance of Parser
        if len(bases) == 1:
            strbase = bases[0].__name__
            if strbase not in _MetaBasicParser:
                raise TypeError("metaclass of %s not found"
                                % bases[0].__name__)
            # we inherit from an already constructed parser, so get metaclass
            clsbase = _MetaBasicParser[strbase]
            # inherit rules from parser
            if hasattr(clsbase, '_rules'):
                cls._rules = clsbase._rules.new_child()
            # inherit hooks from parser
            if hasattr(clsbase, '_hooks'):
                cls._hooks = clsbase._hooks.new_child()
        # add localy defined rules
        if '_rules' in namespace:
            cls._rules.update(namespace['_rules'])
        # add localy defined hooks
        if '_hooks' in namespace:
            cls._hooks.update(namespace['_hooks'])
        # store in global registry
        _MetaBasicParser[name] = cls
        return cls


class BasicParser(metaclass=MetaBasicParser):
    """Emtpy basic parser, contains no rule nor hook.

    Unless you know what you are doing, use Parser instead of this class.

    """
    _rules = collections.ChainMap()
    _hooks = collections.ChainMap()

    def __init__(self, content: str='', stream_name: str="root"):
        self._ignores = [BasicParser.ignore_blanks]
        self.__streams = [Stream(content, stream_name)]
        self.__tags = {}
        self.rulenodes = collections.ChainMap()
        self._lastIgnoreIndex = 0
        self._lastIgnore = False

### READ ONLY @property

    @property
    def _stream(self) -> Stream:
        """The current Stream."""
        return self.__streams[-1]

    @property
    def rules(self) -> dict:
        """
        Return the grammar dict
        """
        return self._rules

### Rule Nodes

    def push_rule_nodes(self) -> bool:
        """Push context variable to store rule nodes."""
        self.rulenodes = self.rulenodes.new_child()
        return True

    def pop_rule_nodes(self) -> bool:
        """Pop context variable that store rule nodes"""
        self.rulenodes = self.rulenodes.parents
        return True

### STREAM

    def parsed_stream(self, sNewStream: str, sName="string"):
        """Push a new Stream into the parser.
        All subsequent called functions will parse this new stream,
        until the 'popStream' function is called.
        """
        self.__streams.append(Stream(sNewStream, sName))

    def pop_stream(self):
        """Pop the last Stream pushed on to the parser stack."""
        self.__streams.pop()

### VARIABLE PRIMITIVES

# TODO(iopi): change for beginTag,endTag,getTag for multicapture,
# and typesetting at endTag (i.e: readCChar,readCString need transcoding)
    def begin_tag(self, name: str) -> Node:
        """Save the current index under the given name."""
        self.__tags[name] = {'begin': self._stream.index}
        return Node(True)

    def end_tag(self, name: str) -> Node:
        """Extract the string between saved and current index."""
        self.__tags[name]['end'] = self._stream.index
        return Node(True)

    def get_tag(self, name: str) -> str:
        """Extract the string previously saved."""
        begin = self.__tags[name]['begin']
        end = self.__tags[name]['end']
        return self._stream[begin:end]

####

    @classmethod
    def set_rules(cls, rules: dict) -> bool:
        """
        Merge internal rules set with the given rules
        """
        cls._rules = cls._rules.new_child()
        for rule_name, rule_pt in rules.items():
            if '.' not in rule_name:
                rule_name = cls.__module__ \
                    + '.' + cls.__name__ \
                    + '.' + rule_name
            meta.set_one(cls._rules, rule_name, rule_pt)
        return True

    @classmethod
    def set_hooks(cls, hooks: dict) -> bool:
        """
        Merge internal hooks set with the given hooks
        """
        cls._hooks = cls._hooks.new_child()
        for hook_name, hook_pt in hooks.items():
            if '.' not in hook_name:
                hook_name = cls.__module__ \
                    + '.' + cls.__name__ \
                    + '.' + hook_name
            meta.set_one(cls._hooks, hook_name, hook_pt)
        return True

    @classmethod
    def set_directives(cls, directives: dict) -> bool:
        """
        Merge internal directives set with the given directives.
        For working directives, attach it only in the dsl.Parser class
        """
        meta._directives = meta._directives.new_child()
        for dir_name, dir_pt in directives.items():
            meta.set_one(meta._directives, dir_name, dir_pt)
            dir_pt.ns_name = dir_name
        return True

    def eval_rule(self, name: str) -> Node:
        """Evaluate a rule by name."""
        self.push_rule_nodes()
        # create a slot value for the result
        return_node = Node()
        import weakref
        self.rulenodes['_'] = weakref.proxy(return_node)
        if name not in self.__class__._rules:
            error.throw("Unknown rule : %s" % name, self)
        res = self.__class__._rules[name](self)
        if res:
            res = return_node
        self.pop_rule_nodes()
        return res

    def eval_hook(self, name: str, ctx: list) -> Node:
        """Evaluate the hook by its name"""
        if name not in self.__class__._hooks:
            # TODO: don't always throw error, could have return True by default
            error.throw("Unknown hook : %s" % name, self)
        return self.__class__._hooks[name](self, *ctx)

### PARSING PRIMITIVES

    def peek_char(self, c: str) -> bool:
        if self.read_eof():
            return False
        return self._stream.peek_char == c

    def peek_text(self, text: str) -> bool:
        """Same as readText but doesn't consume the stream."""
        start = self._stream.index
        stop = start + len(text)
        if stop > self._stream.eos_index:
            return False
        return self._stream[self._stream.index:stop] == text

    def read_char(self, c: str) -> bool:
        """
        Consume the c head byte, increment current index and return True
        else return False. It use peekchar and it's the same as '' in BNF.
        """
        if self.read_eof():
            return False
        self._stream.save_context()
        if c == self._stream.peek_char:
            self._stream.incpos()
            return self._stream.validate_context()
        return self._stream.restore_context()

    def read_until(self, c: str, inhibitor='\\') -> bool:
        """
        Consume the stream while the c byte is not read, else return false
        ex : if stream is " abcdef ", read_until("d"); consume "abcd".
        """
        if self.read_eof():
            return False
        self._stream.save_context()
        while not self.read_eof():
            if self._stream.peek_char == inhibitor:
                # Delete inhibitor and inhibited character
                self._stream.incpos()
                self._stream.incpos()
            if self._stream.peek_char == c:
                self._stream.incpos()
                return self._stream.validate_context()
            self._stream.incpos()
        return self._stream.restore_context()

    def read_until_eof(self) -> bool:
        """Consume all the stream. Same as EOF in BNF."""
        if self.read_eof():
            return True
        self._stream.save_context()
        while not self.read_eof():
            self._stream.incpos()
        return self._stream.validate_context()

    def read_text(self, text: str) -> bool:
        """
        Consume a strlen(text) text at current position in the stream
        else return False.
        Same as "" in BNF
        ex : read_text("ls");.
        """
        if self.read_eof():
            return False
        self._stream.save_context()
        if self.peek_text(text):
            nLength = len(text)
            for _ in range(0, nLength):
                self._stream.incpos()
            return self._stream.validate_context()
        return self._stream.restore_context()

    def read_range(self, begin: str, end: str) -> int:
        """
        Consume head byte if it is >= begin and <= end else return false
        Same as 'a'..'z' in BNF
        """
        if self.read_eof():
            return False
        c = self._stream.peek_char
        if begin <= c <= end:
            self._stream.incpos()
            return True
        return False

### IGNORE CONVENTION

    def ignore_null(self) -> bool:
        """
        Empty ignore convention for notignore
        """
        return True

    def ignore_blanks(self) -> bool:
        """Consume whitespace characters."""
        self._stream.save_context()
        if not self.read_eof() and self._stream.peek_char in " \t\f\r\n":
            while not self.read_eof() and self._stream.peek_char in " \t\f\r\n":
               self._stream.incpos()
            return self._stream.validate_context()
        return self._stream.validate_context()

    def push_ignore(self, ignoreConvention) -> bool:
        """Set the ignore convention"""
        self._ignores.append(ignoreConvention)
        return True

    def pop_ignore(self) -> bool:
        """Remove the last ignore convention"""
        self._ignores.pop()
        return True

    def skip_ignore(self) -> bool:
        self._lastIgnoreIndex = self._stream.index
        if len(self._ignores) > 0:
            self._ignores[-1](self)
        self._lastIgnore = (self._stream.index != self._lastIgnoreIndex)
        return True

    def undo_ignore(self) -> bool:
        # TODO(iopi): wrong don't work in all case
        if self._lastIgnore:
            self._stream.decpos(self._stream.index - self._lastIgnoreIndex)
            self._lastIgnoreIndex = self._stream.index
        return True


class Parser(BasicParser):
    """An ascii parsing primitive library."""
    pass

### BASE RULES


@meta.rule(BasicParser, "Base.read_char")
def read_one_char(self) -> bool:
    """Read one byte in stream"""
    if self.read_eof():
        return False
    self._stream.incpos()
    return True


@meta.rule(BasicParser, "Base.eof")
def read_eof(self) -> bool:
    """Returns true if reached end of the stream."""
    return self._stream.index == self._stream.eos_index


@meta.rule(Parser, "Base.eol")
def read_eol(self) -> bool:
    """Return True if the parser can consume an EOL byte sequence."""
    if self.read_eof():
        return False
    self._stream.save_context()
    self.read_char('\r')
    if self.read_char('\n'):
        return self._stream.validate_context()
    return self._stream.restore_context()


@meta.rule(Parser, "Base.num")
def read_integer(self) -> bool:
    """
    Read following BNF rule else return False
    readInteger ::= ['0'..'9']+ ;
    """
    if self.read_eof():
        return False
    self._stream.save_context()
    c = self._stream.peek_char
    if c.isdigit():
            self._stream.incpos()
            while not self.read_eof():
                c = self._stream.peek_char
                if not c.isdigit():
                    break
                self._stream.incpos()
            return self._stream.validate_context()
    return self._stream.restore_context()

# `Base.id`
@meta.rule(Parser, "Base.id")
def read_identifier(self) -> bool:
    """
    Read following BNF rule else return False
    readIdentifier ::= ['a'..'z'|'A'..'Z'|'_']['0'..'9'|'a'..'z'|'A'..'Z'|'_']* ;
    """
    if self.read_eof():
        return False
    self._stream.save_context()
    c = self._stream.peek_char
    if c.isalpha() or c == '_':
            self._stream.incpos()
            while not self.read_eof():
                c = self._stream.peek_char
                if not (c.isalpha() or c.isdigit() or c == '_'):
                    break
                self._stream.incpos()
            return self._stream.validate_context()
    return self._stream.restore_context()


@meta.rule(Parser, "Base.string")
def read_cstring(self) -> bool:
    """
    Read following BNF rule else return False
    '"' -> ['/'| '"']
    """
    self._stream.save_context()
    idx = self._stream.index
    if self.read_char('\"') and self.read_until('\"', '\\'):
        txt = self._stream[idx:self._stream.index]
        res = Node(self._stream.validate_context())
        return res
    return self._stream.restore_context()


@meta.rule(Parser, "Base.char")
def read_cchar(self) -> bool:
    # TODO(iopi): octal digit, hex digit
    """
    Read following BNF rule else return False
    "'" -> [~"/" "'"]
    """
    self._stream.save_context()
    idx = self._stream.index
    if self.read_char('\'') and self.read_until('\'', '\\'):
        txt = self._stream[idx:self._stream.index]
        res = Node(self._stream.validate_context())
        return res
    return self._stream.restore_context()



from pyrser.parsing.parserTree import ParserTree
from pyrser.parsing.parserTree import Call, CallTrue
from pyrser.parsing.parserTree import Complement, LookAhead, Neg, Until
from pyrser.parsing.parserTree import Hook, Rule, Directive, DirectiveWrapper
from pyrser.parsing.parserTree import Alt, Seq
from pyrser.parsing.parserTree import Rep0N, Rep1N, RepOptional
from pyrser.parsing.parserTree import Capture, Scope
from pyrser.parsing.parserTree import Error
from pyrser.parsing.parserBase import BasicParser, Parser, MetaBasicParser
from pyrser.parsing.parserStream import Stream
from pyrser.parsing.node import Node


__all__ = [
    'Alt',
    'BasicParser',
    'Call',
    'CallTrue',
    'Capture',
    'Complement',
    'Directive',
    'DirectiveWrapper',
    'Error',
    'Hook',
    'LookAhead',
    'Neg',
    'Node',
    'ParserTree',
    'Rule',
    'Rep0N',
    'Rep1N',
    'RepOptional',
    'Scope',
    'Seq',
    'Stream',
    'Until'
]



from pyrser import meta, parsing

@meta.rule(parsing.Parser, "Base.ignore_cxx")
def ignore_cxx(self) -> bool:
    """Consume comments and whitespace characters."""
    self._stream.save_context()
    while not self.read_eof():
        idxref = self._stream.index
        if self._stream.peek_char in " \t\f\r\n":
            while not self.read_eof() and self._stream.peek_char in " \t\f\r\n":
               self._stream.incpos()
        if self.peek_text("//") and self.read_until("\n", ""):
            break
        if self.peek_text("/*"):
            while not self.read_eof() and not self.peek_text("*/"):
                self._stream.incpos()
            if self.read_eof() or not self.read_text("*/"):
                return self._stream.restore_context()
        if idxref == self._stream.index:
            break
    return self._stream.validate_context()

@meta.directive("ignore")
class Ignore(parsing.DirectiveWrapper):
    def begin(self, parser, convention: str):
        if convention == "null":
            parser.push_ignore(parsing.Parser.ignore_null)
        if convention == "C/C++":
            parser.push_ignore(parsing.Parser.ignore_cxx)
        if convention == "blanks":
            parsing.push_ignore(parsing.Parser.ignore_blanks)
        return True

    def end(self, parser, convention: str):
        parser.pop_ignore()
        return True



from pyrser import meta
from pyrser.parsing.parserBase import BasicParser


@meta.hook(BasicParser, "copy")
def copy_node(self, dst, src):
    """
        Basically copy one node to another

        usefull to transmit a node from a terminal
        rule as result of the current rule.

        example:

            R ::=
                In : node #copy(_, node)
            ;

        here the node return by the rule In is
        also the node return by the rule R
    """
    dst.set(src)
    return True



import collections


"""An immutable position in a Stream.

index is the index in the Stream.
lineno is the line number of the position.
col_offset is the column offset of the position on the line.
"""
Position = collections.namedtuple('Position', 'index lineno col_offset')


class Cursor:
    """A mutable position in a Stream.

    It can be initialized or set from an immutable Position.
    """
    def __init__(self, position: Position=Position(0, 1, 1)):
        self.__maxindex = self.__index = position.index
        self.__maxcol = self.__col_offset = position.col_offset
        self.__maxline = self.__lineno = position.lineno
        self.__eol = []

    @property
    def index(self) -> int:
        """The current index of the cursor."""
        return self.__index

    @property
    def lineno(self) -> int:
        """The current line number of the cursor."""
        return self.__lineno

    @property
    def col_offset(self) -> int:
        """The current column offset of the cursor."""
        return self.__col_offset

    @property
    def position(self) -> Position:
        """The current position of the cursor."""
        return Position(self.__index, self.__lineno, self.__col_offset)

    @position.setter
    def position(self, position: Position):
        self.__index = position.index
        self.__lineno = position.lineno
        self.__col_offset = position.col_offset

    @property
    def max_readed_position(self) -> Position:
        """The index of the deepest character readed."""
        return Position(self.__maxindex, self.__maxline, self.__maxcol)

    def step_next_char(self):
        """Puts the cursor on the next character."""
        self.__index += 1
        self.__col_offset += 1
        if self.__index > self.__maxindex:
            self.__maxindex = self.__index
            self.__maxcol = self.__col_offset
            self.__maxline = self.__lineno

    def step_prev_char(self):
        """Puts the cursor on the previous character."""
        self.__col_offset -= 1
        self.__index -= 1

    def step_next_line(self):
        """Sets cursor as begin of next line."""
        self.__eol.append(self.position)
        self.__lineno += 1
        self.__col_offset = 0

    def step_prev_line(self):
        """Sets cursor as end of previous line."""
        #TODO(bps): raise explicit error for unregistered eol
        assert self.__eol[-1].index == self.__index
        self.position = self.__eol.pop()


class Stream:
    """Helps keep track of stream processing progress."""
    def __init__(self, content='', name='stream'):
        self.__content = content
        self.__len = len(content)
        self.__name = name
        self._contexts = []
        self._cursor = Cursor()

    def __len__(self) -> int:
        return self.__len

    def __getitem__(self, key: int or slice) -> str:
        return self.__content.__getitem__(key)

    @property
    def name(self) -> int:
        """Name of the stream."""
        return self.__name

    @property
    def eos_index(self) -> int:
        """End Of Stream index."""
        return self.__len

    @property
    def index(self) -> int:
        """The current position index."""
        return self._cursor.position.index

    @property
    def lineno(self) -> int:
        """The current position line number."""
        return self._cursor.lineno

    @property
    def col_offset(self) -> int:
        """The current position column offset."""
        return self._cursor.col_offset

    @property
    def peek_char(self) -> str:
        """The current position character value."""
        return self.__content[self._cursor.index]

    @property
    def last_readed_line(self) -> str:
        """Usefull string to compute error message."""
        mpos = self._cursor.max_readed_position
        mindex = mpos.index
        # search last \n
        prevline = mindex - 1 if mindex == self.eos_index else mindex
        while prevline >= 0 and self.__content[prevline] != '\n':
            prevline -= 1
        # search next \n
        nextline = mindex
        while nextline < self.eos_index and self.__content[nextline] != '\n':
            nextline += 1
        last_line = self.__content[prevline + 1:nextline]
        return last_line

    def incpos(self, length: int=1) -> int:
        """Increment the cursor to the next character."""
        if length <= 0:
            raise ValueError("length must be positive")
        for _ in range(length):
            if self._cursor.index < self.__len:
                if self.peek_char == '\n':
                    self._cursor.step_next_line()
                self._cursor.step_next_char()
        return self._cursor.index

    def decpos(self, length: int=1) -> int:
        if length <= 0:
            raise ValueError("length must be positive")
        for _ in range(length):
            if 0 < self._cursor.index:
                self._cursor.step_prev_char()
                if self.peek_char == '\n':
                    self._cursor.step_prev_line()
        return self._cursor.index

    def save_context(self) -> bool:
        """Save current position."""
        self._contexts.append(self._cursor.position)
        return True

    def restore_context(self) -> bool:
        """Rollback to previous saved position."""
        self._cursor.position = self._contexts.pop()
        return False

    def validate_context(self) -> bool:
        """Discard previous saved position."""
        del self._contexts[-1]
        return True



# error handling


#forward declaration of BasicParser
class BasicParser:
    pass


class ParseError(Exception):
    def __init__(self, message, stream_name="", pos=None, line="", **kwargs):
        msg = (message + " in {stream_name} at line {line} col {col}\n"
               "{last_read_line}\n"
               "{underline}")
        underline = "%s^" % ('-' * (pos.col_offset - 1))
        kwargs.update(
            message=message, stream_name=stream_name,
            line=pos.lineno, col=pos.col_offset,
            last_read_line=line, underline=underline)
        self.raw_msg = message
        self.msg = msg.format(**kwargs)
        Exception.__init__(self, self.msg)
        self.stream_name = stream_name
        self.error_position = pos
        self.error_line = line


def throw(msg: str, parser: BasicParser, **kw):
    """Convenient function to raise a ParseError"""
    kw.update(
        stream_name=parser._stream.name,
        pos=parser._stream._cursor.max_readed_position,
        line=parser._stream.last_readed_line)
    raise ParseError(msg, **kw)



#: Some meta prog stuff
import functools
import inspect
import collections
import copy


def enum(*sequential, **named):
    """
    Build an enum statement
    """
    #: build enums from parameter
    enums = dict(zip(sequential, range(len(sequential))), **named)
    enums['map'] = copy.copy(enums)
    #: build reverse mapping
    enums['rmap'] = {}
    for key, value in enums.items():
        if type(value) is int:
            enums['rmap'][value] = key
    return type('Enum', (), enums)


#From PEP 362: http://www.python.org/dev/peps/pep-0362/
def checktypes(func):
    """Decorator to verify arguments and return types."""
    sig = inspect.signature(func)

    types = {}
    for param in sig.parameters.values():
        # Iterate through function's parameters and build the list of
        # arguments types
        param_type = param.annotation
        if param_type is param.empty or not inspect.isclass(param_type):
            # Missing annotation or not a type, skip it
            continue

        types[param.name] = param_type

        # If the argument has a type specified, let's check that its
        # default value (if present) conforms with the type.
        if (param.default is not param.empty and
                not isinstance(param.default, param_type)):
            raise ValueError(
                "{func}: wrong type of a default value for {arg!r}".format(
                func=func.__qualname__, arg=param.name))

    def check_type(sig, arg_name, arg_type, arg_value):
        # Internal function that encapsulates arguments type checking
        if not isinstance(arg_value, arg_type):
            raise ValueError("{func}: wrong type of {arg!r} argument, "
                             "{exp!r} expected, got {got!r}".
                             format(func=func.__qualname__, arg=arg_name,
                                    exp=arg_type.__name__,
                                    got=type(arg_value).__name__))

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # Let's bind the arguments
        ba = sig.bind(*args, **kwargs)
        for arg_name, arg in ba.arguments.items():
            # And iterate through the bound arguments
            try:
                type_ = types[arg_name]
            except KeyError:
                continue
            else:
                # OK, we have a type for the argument, lets get the
                # corresponding parameter description from the signature object
                param = sig.parameters[arg_name]
                if param.kind == param.VAR_POSITIONAL:
                    # If this parameter is a variable-argument parameter,
                    # then we need to check each of its values
                    for value in arg:
                        check_type(sig, arg_name, type_, value)
                elif param.kind == param.VAR_KEYWORD:
                    # If this parameter is a variable-keyword-argument
                    # parameter:
                    for subname, value in arg.items():
                        check_type(sig, arg_name + ':' + subname, type_, value)
                else:
                    # And, finally, if this parameter a regular one:
                    check_type(sig, arg_name, type_, arg)

        result = func(*ba.args, **ba.kwargs)

        # The last bit - let's check that the result is correct
        return_type = sig.return_annotation
        if (return_type is not sig.empty and
                isinstance(return_type, type) and
                not isinstance(result, return_type)):

            raise ValueError(
                '{func}: wrong return type, {exp} expected, got {got}'.format(
                func=func.__qualname__, exp=return_type.__name__,
                got=type(result).__name__))
        return result

    return wrapper


# TODO: could be better in a tool module?
#: Addtototo
def set_one(chainmap, thing_name, callobject):
    """ Add a mapping with key thing_name for callobject in chainmap with
        namespace handling.
    """
    namespaces = reversed(thing_name.split("."))
    lstname = []
    for name in namespaces:
        lstname.insert(0, name)
        strname = '.'.join(lstname)
        chainmap[strname] = callobject


def _get_base_class(cls):
    base = cls
    while base.__bases__[0] is not object:
        base = base.__bases__[0]
    return base


def add_method(cls):
    """Attach a method to a class."""
    def wrapper(f):
        #if hasattr(cls, f.__name__):
        #    raise AttributeError("{} already has a '{}' attribute".format(
        #        cls.__name__, f.__name__))
        setattr(cls, f.__name__, f)
        return f
    return wrapper


def hook(cls, hookname=None):
    """Attach a method to a parsing class and register it as a parser hook.

       The method is registered with its name unless hookname is provided.
    """
    if not hasattr(cls, '_hooks'):
        raise TypeError(
            "%s didn't seems to be a BasicParser subsclasse" % cls.__name__)
    class_hook_list = cls._hooks

    def wrapper(f):
        nonlocal hookname
        add_method(cls)(f)
        if hookname is None:
            hookname = f.__name__
        if '.' not in hookname:
            hookname = '.'.join([cls.__module__, cls.__name__, hookname])
        set_one(class_hook_list, hookname, f)
        return f
    return wrapper


def rule(cls, rulename=None):
    """Attach a method to a parsing class and register it as a parser rule.

       The method is registered with its name unless rulename is provided.
    """
    if not hasattr(cls, '_rules'):
        raise TypeError(
            "%s didn't seems to be a BasicParser subsclasse" % cls.__name__)
    class_rule_list = cls._rules

    def wrapper(f):
        nonlocal rulename
        add_method(cls)(f)
        if rulename is None:
            rulename = f.__name__
        if '.' not in rulename:
            rulename = cls.__module__ + '.' + cls.__name__ + '.' + rulename
        set_one(class_rule_list, rulename, f)
        return f
    return wrapper


# module variable for DirectiveWrapper registering
_directives = collections.ChainMap()


def directive(directname=None):
    """Attach a class to a parsing class and register it as a parser directive.

        The class is registered with its name unless directname is provided.
    """
    global _directives
    class_dir_list = _directives

    def wrapper(f):
        nonlocal directname
        if directname is None:
            directname = f.__name__
        f.ns_name = directname
        set_one(class_dir_list, directname, f)
        return f
    return wrapper



from pyrser import dsl
from pyrser import parsing
from pyrser import meta
from pyrser import error
from collections import ChainMap


class MetaGrammar(parsing.MetaBasicParser):
    """Metaclass for all grammars."""
    def __new__(metacls, name, bases, namespace):
        # for multi heritance we have a simple inheritance relation
        # from the first class in declaration order.
        metabp = parsing.MetaBasicParser
        if len(bases) <= 1:
            cls = metabp.__new__(metacls, name, bases, namespace)
        else:
            b = tuple([bases[0]])
            cls = metabp.__new__(metacls, name, b, namespace)
        # lookup for the metaclass of parsing.
        # Grammar magically inherit rules&hooks from Parser
        if 'Parser' in parsing.parserBase._MetaBasicParser:
            clsbase = parsing.parserBase._MetaBasicParser['Parser']
            # link rules&hooks
            cls._rules = clsbase._rules.new_child()
            cls._hooks = clsbase._hooks.new_child()
            # add rules from DSL
            if 'grammar' in namespace and namespace['grammar'] is not None:
                dsl_object = cls.dsl_parser(namespace['grammar'])
                # namespace rules with module/classe name
                for rule_name, rule_pt in dsl_object.get_rules().items():
                    if '.' not in rule_name:
                        rule_name = cls.__module__ \
                            + '.' + cls.__name__ \
                            + '.' + rule_name
                    meta.set_one(cls._rules, rule_name, rule_pt)
            # add localy define rules (and thus overloads)
            if '_rules' in namespace and namespace['_rules'] is not None:
                cls._rules.update(namespace['_rules'])
            # add localy define hooks
            if '_hooks' in namespace and namespace['_hooks'] is not None:
                cls._hooks.update(namespace['_hooks'])
        # Manage Aggregation
        if len(bases) > 1:
            aggreg_rules = ChainMap()
            aggreg_hooks = ChainMap()
            for subgrammar in bases:
                if hasattr(subgrammar, '_rules'):
                    aggreg_rules = ChainMap(*(aggreg_rules.maps
                                            + subgrammar._rules.maps))
                if hasattr(subgrammar, '_hooks'):
                    aggreg_hooks = ChainMap(*(aggreg_hooks.maps
                                            + subgrammar._hooks.maps))
            # aggregate at toplevel the branch grammar
            cls._rules = ChainMap(*(cls._rules.maps + aggreg_rules.maps))
            cls._hooks = ChainMap(*(cls._hooks.maps + aggreg_hooks.maps))
            # clean redondant in chain for rules
            orderedunique_rules = []
            tocpy_rules = set([id(_) for _ in cls._rules.maps])
            for ch in cls._rules.maps:
                idch = id(ch)
                if idch in tocpy_rules:
                    orderedunique_rules.append(ch)
                    tocpy_rules.remove(idch)
            cls._rules = ChainMap(*orderedunique_rules)
            # clean redondant in chain for hooks
            orderedunique_hooks = []
            tocpy_hooks = set([id(_) for _ in cls._hooks.maps])
            for ch in cls._hooks.maps:
                idch = id(ch)
                if idch in tocpy_hooks:
                    orderedunique_hooks.append(ch)
                    tocpy_hooks.remove(idch)
            cls._hooks = ChainMap(*orderedunique_hooks)
        return cls


class Grammar(parsing.Parser, metaclass=MetaGrammar):
    """
    Base class for all grammars.

    This class turn any class A that inherit it into a grammar.
    Taking the description of the grammar in parameter it will add
    all what is what is needed for A to parse it.
    """
    # Text grammar to generate parsing rules for this class.
    grammar = None
    # Name of the default rule to parse the grammar.
    entry = None
    # DSL parsing class
    dsl_parser = dsl.EBNF

    def after_parse(self, node):
        return node

    def parse(self, source=None, entry=None):
        """Parse source using the grammar"""
        if source is not None:
            self.parsed_stream(source)
        if entry is None:
            entry = self.entry
        if entry is None:
            raise ValueError("No entry rule name defined for {}".format(
                self.__class__.__name__))
        res = self.eval_rule(entry)
        if not res:
            raise error.ParseError(
                "Parse error with the rule {rule!r}",
                stream_name=self._stream.name,
                rule=entry,
                pos=self._stream._cursor.max_readed_position,
                line=self._stream.last_readed_line)
        return self.after_parse(res)

    def parse_file(self, filename: str, entry=None):
        """Parse filename using the grammar"""
        import os.path
        if os.path.exists(filename):
            f = open(filename, 'r')
            self.parsed_stream(f.read(), os.path.abspath(filename))
            f.close()
        if entry is None:
            entry = self.entry
        if entry is None:
            raise ValueError("No entry rule name defined for {}".format(
                self.__class__.__name__))
        res = self.eval_rule(entry)
        if not res:
            raise error.ParseError(
                "Parse error with the rule {rule!r}",
                stream_name=self._stream.name,
                rule=entry,
                pos=self._stream._cursor.max_readed_position,
                line=self._stream.last_readed_line)
        return self.after_parse(res)


generated_class = 0


def from_string(bnf: str, *optional_inherit):
    """
    Create a Grammar from a string
    """
    global generated_class
    class_name = "gen_class_" + str(generated_class)
    generated_class += 1
    inherit = [Grammar] + list(optional_inherit)
    scope = {'grammar': bnf, 'entry': None}
    return type(class_name, tuple(inherit), scope)


def from_file(fn: str, *optional_inherit):
    """
    Create a Grammar from a file
    """
    import os.path
    if os.path.exists(fn):
        f = open(fn, 'r')
        bnf = f.read()
        f.close()
        return from_string(bnf, *optional_inherit)
    raise Exception("File not Found!")



from pyrser import parsing
from pyrser import meta
from pyrser import error
from pyrser.directives import ignore

##: TEST

class EBNF(parsing.Parser):
    """
    Basic class for BNF DSL PARSING.
    
    A full parser for the BNF is provided by this class.
    We construct a tree of parserTree to represents, thru functors, BNF semantics.
    """

    def get_rules(self) -> parsing.Node:
        """
        Parse the DSL and provide a dictionnaries of all resulting rules.
        Call by the MetaGrammar class.
        """
        res = self.eval_rule('bnf_dsl')
        if not res:
            raise error.ParseError(
                "Parse error with the rule {rule!r}",
                stream_name=self._stream.name,
                rule='bnf_dsl',
                pos=self._stream._cursor.max_readed_position,
                line=self._stream.last_readed_line)
        return res

    def __init__(self, stream=''):
        """
        Define the DSL parser.
        """
        super().__init__(stream)
        self.set_rules({
            #
            # bnf_dsl ::= @ignore("C/C++") bnf_stmts
            # ;
            #
            'bnf_dsl': parsing.Seq(
                # parserTree is not already construct but Directive need it
                # forward it thru a lambda
                parsing.Directive(ignore.Ignore(),
                                  [("C/C++", str)],
                                  lambda parser: self.__class__._rules['bnf_stmts'](parser)),
            ),

            #
            # bnf_stmts ::= [rule : r #add_rules(_, r) ]+ Base.eof
            # ;
            #
            'bnf_stmts': parsing.Seq(
                parsing.Rep1N(parsing.Seq(
                    parsing.Capture("r", parsing.Rule('rule')),
                    parsing.Hook('add_rules', [("_", parsing.Node),
                                               ("r", parsing.Node)])
                )),
                parsing.Rule('Base.eof')
            ),

            #
            # rule ::= ns_name : rn "::=" alternatives : alts
            #                             #add_rule(_, rn, alts) ';'
            # ;
            #
            'rule': parsing.Seq(
                parsing.Capture("rn", parsing.Rule('ns_name')),
                parsing.Alt(
                    parsing.Call(parsing.Parser.read_text, "::="),
                    parsing.Error("Expected '::='")),
                parsing.Capture("alts", parsing.Rule('alternatives')),
                parsing.Hook('add_rule', [("_", parsing.Node),
                                          ("rn", parsing.Node),
                                          ("alts", parsing.Node)]),
                parsing.Alt(
                    parsing.Call(parsing.Parser.read_char, ';'),
                    parsing.Error("Expected ';'"))
            ),

            #
            # alternatives ::= sequences : alt #add_alt(_, alt)
            #                  ['|' sequences : alt
            #                   #add_alt(_, alt) ]*
            # ;
            #
            'alternatives': parsing.Seq(
                parsing.Capture('alt', parsing.Rule('sequences')),
                parsing.Hook('add_alt', [("_", parsing.Node),
                                         ("alt", parsing.Node)]),
                parsing.Rep0N(
                    parsing.Seq(
                        parsing.Call(parsing.Parser.read_char, '|'),
                        parsing.Capture('alt', parsing.Rule('sequences')),
                        parsing.Hook('add_alt',
                                     [("_", parsing.Node),
                                      ("alt", parsing.Node)])
                    )
                )
            ),

            #
            # sequences ::= [ sequence : cla #add_sequences(_, cla) ]+
            #;
            #
            'sequences': parsing.Rep1N(
                parsing.Seq(
                    parsing.Capture('cla', parsing.Rule('sequence')),
                    parsing.Hook('add_sequences',
                                 [("_", parsing.Node),
                                  ("cla", parsing.Node)])
                )
            ),

            #
            # sequence ::=
            #     [ '~' | "!!" | '!' | "->" ]?: mod
            #     [ ns_name : rid #add_ruleclause_name(_, rid)
            #       | Base.string : txt #add_text(_, txt)
            #       | Base.char : begin ".." Base.char : end
            #         #add_range(_, begin, end)
            #       | Base.char : c #add_char(_, c)
            #       | '[' alternatives : subsequence ']'
            #         #add_subsequence(_, subsequence)
            #     ] #add_mod(_, mod)
            #     [repeat : rpt #add_rpt(_, mod, rpt) ]?
            #     [':' Base.id : cpt #add_capture(_, cpt) ]?
            #     | hook : h #add_hook(_, h)
            #     | directive : d sequence : s #add_directive(_, d, s)
            # ;
            #
            'sequence': parsing.Alt(
                parsing.Seq(
                    parsing.Capture(
                        'mod',
                        parsing.RepOptional(
                            parsing.Alt(
                                parsing.Call(parsing.Parser.read_char, '~'),
                                parsing.Call(parsing.Parser.read_text, '!!'),
                                parsing.Call(parsing.Parser.read_char, '!'),
                                parsing.Call(parsing.Parser.read_text, '->')))),
                    parsing.Alt(
                        parsing.Seq(
                            parsing.Capture('rid', parsing.Rule('ns_name')),
                            parsing.Hook('add_ruleclause_name',
                                         [("_", parsing.Node),
                                          ("rid", parsing.Node)])
                        ),
                        parsing.Seq(
                            parsing.Capture('txt',
                                            parsing.Rule('Base.string')),
                            parsing.Hook('add_text',
                                         [("_", parsing.Node),
                                          ("txt", parsing.Node)])
                        ),
                        parsing.Seq(
                            parsing.Capture('begin',
                                            parsing.Rule('Base.char')),
                            parsing.Call(parsing.Parser.read_text, ".."),
                            parsing.Capture('end', parsing.Rule('Base.char')),
                            parsing.Hook('add_range',
                                         [("_", parsing.Node),
                                          ("begin", parsing.Node),
                                          ("end", parsing.Node)])
                        ),
                        parsing.Seq(
                            parsing.Capture('c', parsing.Rule('Base.char')),
                            parsing.Hook('add_char',
                                         [("_", parsing.Node),
                                          ("c", parsing.Node)])
                        ),
                        parsing.Seq(
                            parsing.Call(parsing.Parser.read_char, "["),
                            parsing.Alt(
                                parsing.Capture('subsequence',
                                                parsing.Rule('alternatives')),
                                parsing.Error("Expected sequences")),
                            parsing.Alt(
                                parsing.Call(parsing.Parser.read_char, "]"),
                                parsing.Error("Expected ']'")),
                            parsing.Hook('add_subsequence',
                                         [("_", parsing.Node),
                                          ("subsequence", parsing.Node)]),
                        )
                    ),
                    parsing.Hook('add_mod', [("_", parsing.Node),
                                             ("mod", parsing.Node)]),
                    parsing.RepOptional(
                        parsing.Seq(
                            parsing.Capture('rpt', parsing.Rule('repeat')),
                            parsing.Hook('add_rpt',
                                         [("_", parsing.Node),
                                          ("mod", parsing.Node),
                                          ("rpt", parsing.Node)])
                        )
                    ),
                    parsing.RepOptional(
                        parsing.Seq(
                            parsing.Call(parsing.Parser.read_text, ":"),
                            parsing.Capture('cpt', parsing.Rule('Base.id')),
                            parsing.Hook('add_capture',
                                         [('_', parsing.Node),
                                          ('cpt', parsing.Node)])
                        )
                    )
                ),
                parsing.Seq(
                    parsing.Capture('h', parsing.Rule('hook')),
                    parsing.Hook('add_hook', [('_', parsing.Node),
                                              ('h', parsing.Node)])
                ),
                parsing.Seq(
                    parsing.Capture('d', parsing.Rule('directive')),
                    parsing.Capture('s', parsing.Rule('sequence')),
                    parsing.Hook('add_directive', [('_', parsing.Node),
                                                   ('d', parsing.Node),
                                                   ('s', parsing.Node)])
                )
            ),

            # TODO: add directive hooks
            # ns_name ::= @ignore("null") [ Base.id ['.' Base.id]* ]: rid
            #             #add_ruleclause_name(_, rid)
            # ;
            #
            'ns_name': parsing.Seq(
                parsing.Capture(
                    'rid',
                    parsing.Scope(
                        parsing.Call(parsing.Parser.push_ignore,
                                     parsing.Parser.ignore_null),
                        parsing.Call(parsing.Parser.pop_ignore),
                        parsing.Seq(
                            parsing.Rule('Base.id'),
                            parsing.Rep0N(
                                parsing.Seq(
                                    parsing.Call(parsing.Parser.read_text,
                                                 "."),
                                    parsing.Alt(
                                        parsing.Rule('Base.id'),
                                        parsing.Error(
                                            "Expected identifier after '.'"))
                                )
                            )
                        )
                    )
                ),
                parsing.Hook('add_ruleclause_name',
                             [("_", parsing.Node),
                              ("rid", parsing.Node)])
            ),

            #
            # repeat ::= '?' #add_optional(_)
            #          | '*' #add_0N(_)
            #          | '+' #add_1N(_)
            # ;
            #
            'repeat': parsing.Alt(
                parsing.Seq(
                    parsing.Call(parsing.Parser.read_char, '?'),
                    parsing.Hook('add_optional', [("_", parsing.Node)])
                ),
                parsing.Seq(
                    parsing.Call(parsing.Parser.read_char, '*'),
                    parsing.Hook('add_0N', [("_", parsing.Node)])
                ),
                parsing.Seq(
                    parsing.Call(parsing.Parser.read_char, '+'),
                    parsing.Hook('add_1N', [("_", parsing.Node)])
                ),
            ),

            #
            # hook ::= '#' ns_name : n #hook_name(_, n)
            #          ['(' param : p #hook_param(_, p)
            #              [',' param : p #hook_param(_, p)]*
            #           ')']?
            # ;
            #
            'hook': parsing.Seq(
                parsing.Call(parsing.Parser.read_char, '#'),
                parsing.Capture('n', parsing.Rule('ns_name')),
                parsing.Hook('hook_name',
                             [('_', parsing.Node), ('n', parsing.Node)]),
                parsing.RepOptional(
                    parsing.Seq(
                        parsing.Call(parsing.Parser.read_char, '('),
                        parsing.Alt(
                            parsing.Capture('p', parsing.Rule('param')),
                            parsing.Error("Expected parameter")),
                        parsing.Hook('hook_param', [('_', parsing.Node),
                                                    ('p', parsing.Node)]),
                        parsing.Rep0N(
                            parsing.Seq(
                                parsing.Call(parsing.Parser.read_char, ','),
                                parsing.Alt(
                                    parsing.Capture(
                                        'p',
                                        parsing.Rule('param')),
                                    parsing.Error("Expected parameter")),
                                parsing.Hook('hook_param',
                                             [('_', parsing.Node),
                                              ('p', parsing.Node)]))
                        ),
                        parsing.Alt(
                            parsing.Call(parsing.Parser.read_char, ')'),
                            parsing.Error("Expected ')'"))
                    )
                ),
            ),

            #
            # directive ::= '@' ns_name : n #hook_name(_, n)
            #               ['(' param : p #hook_param(_, p)
            #                   [',' param : p #hook_param(_, p)]*
            #                ')']?
            # ;
            'directive': parsing.Seq(
                parsing.Call(parsing.Parser.read_char, '@'),
                parsing.Capture('n', parsing.Rule('ns_name')),
                parsing.Hook('hook_name', [('_', parsing.Node),
                                           ('n', parsing.Node)]),
                parsing.RepOptional(
                    parsing.Seq(
                        parsing.Call(parsing.Parser.read_char, '('),
                        parsing.Alt(
                            parsing.Capture('p', parsing.Rule('param')),
                            parsing.Error("Expected parameter")),
                        parsing.Hook('hook_param',
                                     [('_', parsing.Node),
                                      ('p', parsing.Node)]),
                        parsing.Rep0N(
                            parsing.Seq(
                                parsing.Call(parsing.Parser.read_char, ','),
                                parsing.Alt(
                                    parsing.Capture(
                                        'p',
                                        parsing.Rule('param')),
                                    parsing.Error("Expected parameter")),
                                parsing.Hook('hook_param',
                                             [('_', parsing.Node),
                                              ('p', parsing.Node)]),
                            )
                        ),
                        parsing.Alt(
                            parsing.Call(parsing.Parser.read_char, ')'),
                            parsing.Error("Expected ')'"))
                    )
                ),
            ),

            #
            # param ::= Base.num :n #param_num(_, n)
            #         | Base.string : s #param_str(_, s)
            #         | Base.char : c #param_char(_, c)
            #         | ns_name : i #param_id(_, i)
            # ;
            #
            'param': parsing.Alt(
                parsing.Seq(
                    parsing.Capture('n', parsing.Rule('Base.num')),
                    parsing.Hook('param_num', [('_', parsing.Node),
                                               ('n', parsing.Node)])
                ),
                parsing.Seq(
                    parsing.Capture('s', parsing.Rule('Base.string')),
                    parsing.Hook('param_str', [('_', parsing.Node),
                                               ('s', parsing.Node)])
                ),
                parsing.Seq(
                    parsing.Capture('c', parsing.Rule('Base.char')),
                    parsing.Hook('param_char', [('_', parsing.Node),
                                                ('c', parsing.Node)])
                ),
                parsing.Seq(
                    parsing.Capture('i', parsing.Rule('ns_name')),
                    parsing.Hook('param_id', [('_', parsing.Node),
                                              ('i', parsing.Node)])
                ),
            ),
        })

# Hooks part
# ----------
# All these functions are automatically added to class EBNF


@meta.hook(EBNF, "EBNF.add_mod")
def add_mod(self, seq, mod):
    """Create a parserTree.{Complement, LookAhead, Neg, Until}"""
    if mod.value == '~':
        seq.parser_tree = parsing.Complement(seq.parser_tree)
    elif mod.value == '!!':
        seq.parser_tree = parsing.LookAhead(seq.parser_tree)
    elif mod.value == '!':
        seq.parser_tree = parsing.Neg(seq.parser_tree)
    elif mod.value == '->':
        seq.parser_tree = parsing.Until(seq.parser_tree)
    return True


@meta.hook(EBNF, "EBNF.add_ruleclause_name")
def add_ruleclause_name(self, ns_name, rid) -> bool:
    """Create a parserTree.Rule"""
    ns_name.value = rid.value
    ns_name.parser_tree = parsing.Rule(ns_name.value)
    return True


@meta.hook(EBNF, "EBNF.add_rules")
def add_rules(self, bnf, r) -> bool:
    """Attach a parser tree to the dict of rules"""
    bnf[r.rulename] = r.parser_tree
    return True


@meta.hook(EBNF, "EBNF.add_rule")
def add_rule(self, rule, rn, alts) -> bool:
    """Add the rule name"""
    rule.rulename = rn.value
    rule.parser_tree = alts.parser_tree
    return True


@meta.hook(EBNF, "EBNF.add_sequences")
def add_sequences(self, sequences, cla) -> bool:
    """Create a parserTree.Seq"""
    if not hasattr(sequences, 'parser_tree'):
        # forward sublevel of sequence as is
        sequences.parser_tree = cla.parser_tree
    else:
        oldnode = sequences
        if isinstance(oldnode.parser_tree, parsing.Seq):
            oldpt = list(oldnode.parser_tree.ptlist)
        else:
            oldpt = [oldnode.parser_tree]
        oldpt.append(cla.parser_tree)
        sequences.parser_tree = parsing.Seq(*tuple(oldpt))
    return True


@meta.hook(EBNF, "EBNF.add_alt")
def add_alt(self, alternatives, alt) -> bool:
    """Create a parserTree.Alt"""
    if not hasattr(alternatives, 'parser_tree'):
        # forward sublevel of alt as is
        if hasattr(alt, 'parser_tree'):
            alternatives.parser_tree = alt.parser_tree
        else:
            alternatives.parser_tree = alt
    else:
        oldnode = alternatives
        if isinstance(oldnode.parser_tree, parsing.Alt):
            oldpt = list(oldnode.parser_tree.ptlist)
        else:
            oldpt = [oldnode.parser_tree]
        oldpt.append(alt.parser_tree)
        alternatives.parser_tree = parsing.Alt(*tuple(oldpt))
    return True


@meta.hook(EBNF, "EBNF.add_char")
def add_char(self, sequence, c):
    """Add a read_char primitive"""
    sequence.parser_tree = parsing.Call(parsing.Parser.read_char,
                                        c.value.strip("'"))
    return True


@meta.hook(EBNF, "EBNF.add_text")
def add_text(self, sequence, txt):
    """Add a read_text primitive"""
    sequence.parser_tree = parsing.Call(parsing.Parser.read_text,
                                        txt.value.strip('"'))
    return True


@meta.hook(EBNF, "EBNF.add_range")
def add_range(self, sequence, begin, end):
    """Add a read_range primitive"""
    sequence.parser_tree = parsing.Call(parsing.Parser.read_range,
                                        begin.value.strip("'"),
                                        end.value.strip("'"))
    return True


@meta.hook(EBNF, "EBNF.add_rpt")
def add_rpt(self, sequence, mod, pt):
    """Add a repeater to the previous sequence"""
    if mod.value == '!!':
        error.throw("Cannot repeat a lookahead rule", self)
    if mod.value == '!':
        error.throw("Cannot repeat a negated rule", self)
    oldnode = sequence
    sequence.parser_tree = pt.functor(oldnode.parser_tree)
    return True


@meta.hook(EBNF, "EBNF.add_capture")
def add_capture(self, sequence, cpt):
    """Create a parserTree.Capture"""
    sequence.parser_tree = parsing.Capture(cpt.value, sequence.parser_tree)
    return True


@meta.hook(EBNF, "EBNF.add_subsequence")
def add_subsequence(self, sequence, subsequence):
    """Add a subsequence into a sequence"""
    sequence.parser_tree = subsequence.parser_tree
    return True


@meta.hook(EBNF, "EBNF.add_optional")
def add_optional(self, repeat):
    """Create a parserTree.RepOptional"""
    repeat.functor = parsing.RepOptional
    return True


@meta.hook(EBNF, "EBNF.add_0N")
def add_0N(self, repeat):
    """Create a parserTree.Rep0N"""
    repeat.functor = parsing.Rep0N
    return True


@meta.hook(EBNF, "EBNF.add_1N")
def add_1N(self, repeat):
    """Create a parserTree.Rep1N"""
    repeat.functor = parsing.Rep1N
    return True


@meta.hook(EBNF, "EBNF.add_hook")
def add_hook(self, sequence, h):
    """Create a parserTree.Hook"""
    sequence.parser_tree = parsing.Hook(h.name, h.listparam)
    return True


@meta.hook(EBNF, "EBNF.param_num")
def param_num(self, param, n):
    """Parse a int in parameter list"""
    param.pair = (int(n.value), int)
    return True


@meta.hook(EBNF, "EBNF.param_str")
def param_str(self, param, s):
    """Parse a str in parameter list"""
    param.pair = (s.value.strip('"'), str)
    return True


@meta.hook(EBNF, "EBNF.param_char")
def param_char(self, param, c):
    """Parse a char in parameter list"""
    param.pair = (c.value.strip("'"), str)
    return True


@meta.hook(EBNF, "EBNF.param_id")
def param_id(self, param, i):
    """Parse a node name in parameter list"""
    param.pair = (i.value, parsing.Node)
    return True


@meta.hook(EBNF, "EBNF.hook_name")
def hook_name(self, hook, n):
    """Parse a hook name"""
    hook.name = n.value
    hook.listparam = []
    return True


@meta.hook(EBNF, "EBNF.hook_param")
def hook_param(self, hook, p):
    """Parse a hook parameter"""
    hook.listparam.append(p.pair)
    return True


@meta.hook(EBNF, "EBNF.add_directive")
def add_directive(self, sequence, d, s):
    """Add a directive in the sequence"""
    if d.name not in meta._directives:
        raise TypeError("Unkown directive %s" % d.name)
    the_class = meta._directives[d.name]
    sequence.parser_tree = parsing.Directive(the_class(), d.listparam,
                                             s.parser_tree)
    return True



class Node(dict):
    """Base class for node manipulation."""
    def __init__(self, val=True):
        if type(val) not in (bool, Node):
            raise TypeError(
                "{} is neither a Node nor a Boolean".format(
                    type(val).__name__))
        self._bool = bool(val)

    def __bool__(self):
        return self._bool

    def __repr__(self):
        items = []
        if len(self) > 0:
            items.append(repr(self.items()))
        for k, v in vars(self).items():
            items.append("{} = {}".format(k, repr(v)))
        return "{}({})".format(self.__class__.__name__, ', '.join(items))

    def set(self, othernode):
        """allow to completly mutate the node with another instance"""
        self.__class__ = othernode.__class__
        if len(othernode) > 0:
            for k, v in othernode.items():
                self[k] = v
        for k, v in vars(othernode).items():
            setattr(self, k, v)



import types
from pyrser import meta, error
from pyrser.parsing.parserBase import BasicParser
from pyrser.parsing.node import Node


class ParserTree:
    """Dummy Base class for all parse tree classes.

    common property:
        pt if contain a ParserTree
        ptlist if contain a list of ParserTree
    """
    pass


class Seq(ParserTree):
    """A B C bnf primitive as a functor."""

    def __init__(self, *ptlist: ParserTree):
        ParserTree.__init__(self)
        if len(ptlist) == 0:
            raise TypeError("Expected ParserTree")
        self.ptlist = ptlist

    def __call__(self, parser: BasicParser) -> bool:
        parser._stream.save_context()
        for pt in self.ptlist:
            parser.skip_ignore()
            if not pt(parser):
                return parser._stream.restore_context()
        return parser._stream.validate_context()


class Scope(ParserTree):
    """functor to wrap SCOPE/rule directive or just []."""

    def __init__(self, begin: Seq, end: Seq, pt: Seq):
        ParserTree.__init__(self)
        self.begin = begin
        self.end = end
        self.pt = pt

    def __call__(self, parser: BasicParser) -> Node:
        if not self.begin(parser):
            return False
        res = self.pt(parser)
        if not self.end(parser):
            return False
        return res


class LookAhead(ParserTree):
    """!!A bnf primitive as a functor."""
    def __init__(self, pt: ParserTree):
        ParserTree.__init__(self)
        self.pt = pt

    def __call__(self, parser: BasicParser) -> bool:
        parser._stream.save_context()
        res = self.pt(parser)
        parser._stream.restore_context()
        return res


class Neg(ParserTree):
    """!A bnf primitive as a functor."""

    def __init__(self, pt: ParserTree):
        ParserTree.__init__(self)
        self.pt = pt

    def __call__(self, parser: BasicParser):
        parser._stream.save_context()
        if self.pt(parser):
            res = parser._stream.restore_context()
            return res
        return parser._stream.validate_context()


class Complement(ParserTree):
    """~A bnf primitive as a functor."""

    def __init__(self, pt: ParserTree):
        ParserTree.__init__(self)
        self.pt = pt

    def __call__(self, parser: BasicParser) -> bool:
        if parser.read_eof():
            return False
        ## skip/undo?
        parser.skip_ignore()
        parser._stream.save_context()
        res = self.pt(parser)
        if not res:
            parser._stream.incpos()
            return parser._stream.validate_context()
        parser._stream.restore_context()
        parser.undo_ignore()
        return False


class Until(ParserTree):
    """->A bnf primitive as a functor."""

    def __init__(self, pt: ParserTree):
        ParserTree.__init__(self)
        self.pt = pt

    def __call__(self, parser: BasicParser) -> bool:
        ## skip/undo?
        parser.skip_ignore()
        parser._stream.save_context()
        while not parser.read_eof():
            res = self.pt(parser)
            if res:
                return parser._stream.validate_context()
            parser._stream.incpos()
        parser._stream.restore_context()
        parser.undo_ignore()
        return False


class Call(ParserTree):
    """Functor wrapping a BasicParser method call in a BNF clause."""

    def __init__(self, callObject, *params):
        ParserTree.__init__(self)
        self.callObject = callObject
        self.params = params

    def __call__(self, parser: BasicParser) -> Node:
        return self.callObject(parser, *self.params)


class CallTrue(Call):
    """Functor to wrap arbitrary callable object in BNF clause."""

    def __call__(self, parser: BasicParser) -> Node:
        self.callObject(*self.params)
        return True


class Capture(ParserTree):
    """Functor to handle capture variables."""

    def __init__(self, tagname: str, pt: ParserTree):
        ParserTree.__init__(self)
        if not isinstance(tagname, str) or len(tagname) == 0:
            raise TypeError("Illegal tagname for capture")
        self.tagname = tagname
        self.pt = pt

    def __call__(self, parser: BasicParser) -> Node:
        if parser.begin_tag(self.tagname):
            # create a node for the capture (visible during rule evaluation)
            if self.tagname not in parser.rulenodes.maps[0]:
                parser.rulenodes[self.tagname] = Node()
            res = self.pt(parser)
            if res and parser.end_tag(self.tagname):
                text = parser.get_tag(self.tagname)
                # wrap it in a Node instance
                if type(res) is bool:
                    res = Node(res)
                #TODO(iopi): should be a future capture object for multistream
                # capture
                if not hasattr(res, 'value'):
                    res.value = text
                parser.rulenodes[self.tagname].set(res)
                return res
        return False


class Alt(ParserTree):
    """A | B bnf primitive as a functor."""

    def __init__(self, *ptlist: Seq):
        ParserTree.__init__(self)
        self.ptlist = ptlist

    def __call__(self, parser: BasicParser) -> Node:
        for pt in self.ptlist:
            parser._stream.save_context()
            parser.skip_ignore()
            res = pt(parser)
            if res:
                parser._stream.validate_context()
                return res
            parser._stream.restore_context()
        return False


class RepOptional(ParserTree):
    """[]? bnf primitive as a functor."""
    def __init__(self, pt: Seq):
        ParserTree.__init__(self)
        self.pt = pt

    def __call__(self, parser: BasicParser) -> bool:
        parser.skip_ignore()
        res = self.pt(parser)
        if res:
            return res
        return True


class Rep0N(ParserTree):
    """[]* bnf primitive as a functor."""

    def __init__(self, pt: Seq):
        ParserTree.__init__(self)
        self.pt = pt

    def __call__(self, parser: BasicParser) -> bool:
        parser.skip_ignore()
        parser.push_rule_nodes()
        while self.pt(parser):
            parser.skip_ignore()
        parser.pop_rule_nodes()
        return True


class Rep1N(ParserTree):
    """[]+ bnf primitive as a functor."""

    def __init__(self, pt: Seq):
        ParserTree.__init__(self)
        self.pt = pt

    def __call__(self, parser: BasicParser) -> bool:
        parser._stream.save_context()
        # skip/undo
        parser.skip_ignore()
        parser.push_rule_nodes()
        if self.pt(parser):
            parser.skip_ignore()
            while self.pt(parser):
                parser.skip_ignore()
            parser.pop_rule_nodes()
            return parser._stream.validate_context()
        parser.pop_rule_nodes()
        return parser._stream.restore_context()


class Error(ParserTree):
    """Raise an error."""

    def __init__(self, msg: str, **kwargs):
        self.msg = msg
        self.kw = kwargs

    def __call__(self, parser: BasicParser) -> bool:
        error.throw(self.msg, parser, **self.kw)


class Rule(ParserTree):
    """Call a rule by its name."""

    #TODO(iopi): Handle additionnal value
    def __init__(self, name: str):
        ParserTree.__init__(self)
        self.name = name

    def __call__(self, parser: BasicParser) -> Node:
        return parser.eval_rule(self.name)


class Hook(ParserTree):
    """Call a hook by his name."""

    def __init__(self, name: str, param: [(object, type)]):
        ParserTree.__init__(self)
        self.name = name
        # compose the list of value param, check type
        for v, t in param:
            if type(t) is not type:
                raise TypeError("Must be pair of value and type (i.e: int, "
                                "str, Node)")
        self.param = param

    def __call__(self, parser: BasicParser) -> bool:
        valueparam = []
        for v, t in self.param:
            if t is Node:
                import weakref
                if v not in parser.rulenodes:
                    from pyrser.passes.to_yml import to_yml
                    print(to_yml(parser.rulenodes))
                    error.throw("Unknown capture variable : %s" % v, parser)
                if type(parser.rulenodes[v]) is weakref.ProxyType:
                    valueparam.append(parser.rulenodes[v])
                else:
                    valueparam.append(weakref.proxy(parser.rulenodes[v]))
            elif type(v) is t:
                valueparam.append(v)
            else:
                raise TypeError("Type mismatch expected {} got {}".format(
                    t, type(v)))
        return parser.eval_hook(self.name, valueparam)


class MetaDirectiveWrapper(type):
    """metaclass of all DirectiveWrapper subclasses.
    ensure that begin and end exists in subclasses as method"""
    def __new__(metacls, name, bases, namespace):
        cls = type.__new__(metacls, name, bases, namespace)
        if 'begin' not in namespace:
            raise TypeError(
                "DirectiveWrapper %s must have a begin method" % name)
        if not(isinstance(namespace['begin'], types.FunctionType)):
            raise TypeError(
                "'begin' not a function class in DirectiveWrapper %s" % name)
        if 'end' not in namespace:
            raise TypeError(
                "DirectiveWrapper %s subclasse must have a end method" % name)
        if not(isinstance(namespace['end'], types.FunctionType)):
            raise TypeError(
                "'end' not a function class in DirectiveWrapper %s" % name)
        return cls


class DirectiveWrapper(metaclass=MetaDirectiveWrapper):
    """functor to wrap begin/end directive"""

    def __init__(self, ):
        ParserTree.__init__(self)

    def checkParam(self, params: list):
        if (not hasattr(self.__class__, 'begin') or
                not hasattr(self.__class__, 'end')):
            return False
        pbegin = self.__class__.begin.__code__.co_varnames
        tbegin = self.__class__.begin.__annotations__
        pend = self.__class__.end.__code__.co_varnames
        tend = self.__class__.end.__annotations__
        idx = 0
        for pname in pbegin:
            if pname in tbegin:
                if not isinstance(params[idx], tbegin[pname]):
                    raise TypeError(
                        "{}: Wrong parameter in begin method parameter {} "
                        "expected {} got {}".format(
                            self.__class__.__name__,
                            idx, type(params[idx]),
                            tbegin[pname]))
                idx += 1
        idx = 0
        for pname in pend:
            if pname in tend:
                if not isinstance(params[idx], tend[pname]):
                    raise TypeError(
                        "{}: Wrong parameter in end method parameter {} "
                        "expected {} got {}".format(
                            self.__class__.__name__,
                            idx,
                            type(params[idx]),
                            tend[pname]))
                idx += 1
        return True

    def begin(self):
        pass

    def end(self):
        pass


class Directive(ParserTree):
    """functor to wrap directive HOOKS"""

    def __init__(self, directive: DirectiveWrapper, param: [(object, type)],
                 pt: ParserTree):
        ParserTree.__init__(self)
        self.directive = directive
        self.pt = pt
        # compose the list of value param, check type
        for v, t in param:
            if type(t) is not type:
                raise TypeError(
                    "Must be pair of value and type (i.e: int, str, Node)")
        self.param = param

    def __call__(self, parser: BasicParser) -> Node:
        valueparam = []
        for v, t in self.param:
            if t is Node:
                valueparam.append(weakref.proxy(parser.rulenodes[v]))
            elif type(v) is t:
                valueparam.append(v)
            else:
                raise TypeError(
                    "Type mismatch expected {} got {}".format(t, type(v)))
        if not self.directive.checkParam(valueparam):
            return False
        if not self.directive.begin(parser, *valueparam):
            return False
        res = self.pt(parser)
        if not self.directive.end(parser, *valueparam):
            return False
        return res



from pyrser import meta
from pyrser.parsing.parserBase import BasicParser


@meta.hook(BasicParser, "echo")
def echo_nodes(self, *rest):
    """
    Print nodes

    example:

        R ::=
            In : node #echo(node)
        ;
    """
    print(*rest)
    return True

@meta.hook(BasicParser, "vars")
def vars_nodes(self, rest):
    """
    Vars one node instance

    example:

        R ::=
            In : node #vars(node)
        ;
    """
    print(vars(rest))
    return True




import setuptools


setuptools.setup(
    name='pyrser',
    version='0.0.1',
    url='https://code.google.com/p/pyrser/',
    license='GPLv3',
    maintainer='Lionel Auroux',
    maintainer_email='lionel.auroux@gmail.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.3',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Software Development :: Compilers',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=['pyrser', 'pyrser.directives', 'pyrser.hooks', 'pyrser.parsing'],
#    test_loader='unittest:TestLoader',
#    test_suite='tests'
)
